import numba
import numpy as np

from .utils import pd_to_np


@numba.jit(nopython=True, cache=True)
def _FILTER_1d_nb(arr, n):
    """内部函数，请勿直接调用，请参考"""
    i = 0
    while i < len(arr):
        if arr[i]:
            arr[i + 1:i + 1 + n] = 0
            i += n + 1
    return arr


@numba.jit(nopython=True, cache=True)
def _FILTER_2d_nb(arr, n):
    """内部函数，请勿直接调用"""
    for j in range(arr.shape[1]):
        i = 0
        while i < len(arr):
            if arr[i, j]:
                arr[i + 1:i + 1 + n, j] = 0
                i += n + 1
    return arr


@numba.jit(nopython=True, cache=True)
def _fill_notna_1d_nb(arr, fill_value, n: int):
    k = n
    for i in range(len(arr)):
        if np.isnan(arr[i]):
            continue
        arr[i] = fill_value
        k -= 1
        if k <= 0:
            break


@numba.jit(nopython=True, cache=True)
def _fill_notna_2d_nb(arr, fill_value, n: int):
    for i in range(arr.shape[1]):
        k = n
        for j in range(arr.shape[0]):
            if np.isnan(arr[j, i]):
                continue
            arr[j, i] = fill_value
            k -= 1
            if k <= 0:
                break


def fill_notna(arr, fill_value, n=1):
    """查找第一个非nan,将其改成fill_value"""
    if n < 1:
        return arr
    arr = pd_to_np(arr, copy=True)
    if arr.ndim == 2:
        _fill_notna_2d_nb(arr, fill_value, n)
    else:
        _fill_notna_1d_nb(arr, fill_value, n)
    return arr



"""这里存放因速度慢被淘汰的代码，因为有教学意义而保留"""
import numpy as np
import pandas as pd
import talib as ta

import ta_cn.talib as ta2d
from .ema import EMA_1_PD, WS_SUM, SMA
from .maths import MAX, ABS
from .nb import fill_notna
from .over_bought_over_sold import ROC, TYP
from .reference import HHV, LLV, MA, REF, SUM, TR
from .statistics import AVEDEV
from .utils import np_to_pd


def ATR(high, low, close, timeperiod=14):
    """ATR真实波幅N日平均

    talib的ATR算法类似于SMA，所以要重写此处才与中国ATR相同
    """
    return SMA(TR(high, low, close), timeperiod)


def ATR_CN(high, low, close, timeperiod=14):
    """ATR真实波幅N日平均

    talib的ATR算法类似于EMA，所以要重写此处才与中国ATR相同
    """

    # 以下要慢一些，不采用了
    def func(high, low, close, timeperiod):
        return ta.SMA(ta.TRANGE(high, low, close), timeperiod)

    return ta2d.tafunc_nditer(func, [high, low, close], {'timeperiod': timeperiod}, ['real'])


def MACD_CN(real: pd.DataFrame, fastperiod=12, slowperiod=26, signalperiod=9):
    # 中国区使用公式自行实现的算法，由于多次用到EMA，效率不高，建议直接使用MACDEXT
    DIF = EMA_1_PD(real, fastperiod) - EMA_1_PD(real, slowperiod)
    DEA = EMA_1_PD(DIF, signalperiod)
    MACD = (DIF - DEA) * 2
    return DIF, DEA, MACD


def KDJ_CN(high, low, close, fastk_period=9, M1=3, M2=3):
    """KDJ指标"""
    hh = HHV(high, fastk_period)
    ll = LLV(low, fastk_period)
    RSV = (close - ll) / (hh - ll) * 100
    K = EMA_1_PD(RSV, (M1 * 2 - 1))
    D = EMA_1_PD(K, (M2 * 2 - 1))

    J = K * 3 - D * 2
    return K, D, J


def TRIX_CN(real, timeperiod=12):
    """三重指数平滑均线

    由EMA算法差异导致的不同
    """
    TR = EMA_1_PD(EMA_1_PD(EMA_1_PD(real, timeperiod), timeperiod), timeperiod)
    return ROC(TR, 1)


def CCI(high, low, close, timeperiod=14):
    tp = TYP(high, low, close)
    return (tp - MA(tp, timeperiod)) / (0.015 * AVEDEV(tp, timeperiod))


def RSI(real, timeperiod=24):
    """RSI指标"""
    DIF = real - REF(real, 1)
    return SMA(MAX(DIF, 0), timeperiod, 1) / SMA(ABS(DIF), timeperiod, 1) * 100


def WMA(real, timeperiod=5):
    """加权移动平均"""

    def func(x):
        # 复制于MyTT,比tqsdk中tafunc中计算要快
        return x[::-1].cumsum().sum() * 2 / timeperiod / (timeperiod + 1)

    return np_to_pd(real).rolling(timeperiod).apply(func, raw=True)


def MFI(high, low, close, volume, timeperiod=14):
    """MFI指标"""
    tp = TYP(high, low, close)
    tpv = tp * volume
    # 比TALIB结果多一个数字，通过置空实现与TA-LIB完全一样
    tpv = fill_notna(tpv, fill_value=np.nan, n=1)

    is_raising = tp > REF(tp, 1)
    pos_sum = SUM(is_raising * tpv, timeperiod)
    neg_sum = SUM(~is_raising * tpv, timeperiod)
    return 100 * pos_sum / (pos_sum + neg_sum)


def DM(high, low, timeperiod=14):
    """Directional Movement方向动量

    WS_SUM威尔德平滑求和
    """
    HD = high - REF(high, 1)
    LD = REF(low, 1) - low
    # talib中是用的威尔德平滑
    PDM = WS_SUM(((HD > 0) & (HD > LD)) * HD, timeperiod)
    MDM = WS_SUM(((LD > 0) & (LD > HD)) * LD, timeperiod)
    return PDM, MDM


def DM_CN(high, low, timeperiod=14):
    """中国版DM

    SUM滚动求和
    """
    HD = high - REF(high, 1)
    LD = REF(low, 1) - low
    # 而中国版一般是直接滚动求和
    PDM = SUM(((HD > 0) & (HD > LD)) * HD, timeperiod)
    MDM = SUM(((LD > 0) & (LD > HD)) * LD, timeperiod)
    return PDM, MDM


def DI(high, low, close, timeperiod=14):
    """Directional Indicator方向指标"""
    TRS = WS_SUM(TR(high, low, close), timeperiod)

    # 与talib相比多了一个数字，所以删除一下
    TRS = fill_notna(TRS, fill_value=np.nan, n=1)

    PDM, MDM = DM(high, low, timeperiod)
    PDI = PDM * 100 / TRS
    MDI = MDM * 100 / TRS
    return PDI, MDI


def DI_CN(high, low, close, timeperiod=14):
    """中国版DI

    区别是SUM与WS_SUM
    """
    TRS = SUM(TR(high, low, close), timeperiod)
    PDM, MDM = DM_CN(high, low, timeperiod)
    PDI = PDM * 100 / TRS
    MDI = MDM * 100 / TRS
    return PDI, MDI


def DMI(high, low, close, timeperiod=14):
    """趋向指标"""
    PDI, MDI = DI(high, low, close, timeperiod=timeperiod)
    ADX = SMA(ABS(PDI - MDI) / (PDI + MDI) * 100, timeperiod)
    # 这里timeperiod-1，才正好与talib对应
    ADXR = (ADX + REF(ADX, timeperiod - 1)) / 2
    return PDI, MDI, ADX, ADXR


def DMI_CN(high, low, close, timeperiod=14):
    # DI算法不同
    PDI, MDI = DI_CN(high, low, close, timeperiod=timeperiod)
    # ADX中的MA与SMA不同
    ADX = MA(ABS(PDI - MDI) / (PDI + MDI) * 100, timeperiod)
    # timeperiod与timeperiod-1不同
    ADXR = (ADX + REF(ADX, timeperiod)) / 2
    return PDI, MDI, ADX, ADXR



import talib as ta

TA_COMPATIBILITY_DEFAULT = 0  # 使用MA做第一个值
TA_COMPATIBILITY_METASTOCK = 1  # 使用Price做第一个值

_COMPATIBILITY_ENABLE_ = False


def TA_SET_COMPATIBILITY_ENABLE(enable):
    global _COMPATIBILITY_ENABLE_
    _COMPATIBILITY_ENABLE_ = enable


def TA_SET_COMPATIBILITY(compatibility):
    global _COMPATIBILITY_ENABLE_
    if _COMPATIBILITY_ENABLE_:
        print('do ta.set_compatibility', compatibility)
        ta.set_compatibility(compatibility)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-06-16
"""
Examples
--------
只要替换导入即可直接支持二维矩阵
>>> import talib as ta
>>> import ta_cn.talib as ta

"""
import sys as _sys

import numpy as np
import pandas as pd
import talib as _talib
from talib import abstract as _abstract


def tafunc_nditer(tafunc, args, kwargs, output_names):
    """数据应用函数

    Parameters
    ----------
    tafunc
        计算单列的函数
    output_names: list
        tafunc输出参数名

    Returns
    -------
    tuple
        输出数组

    """

    def is_np_pd(x):
        """是否几种特数类型"""
        return isinstance(x, (pd.DataFrame, pd.DataFrame, np.ndarray))

    def pd_to_np(x):
        """pandas格式转numpy"""
        if isinstance(x, (pd.Series, pd.DataFrame)):
            return x.values
        return x

    # 分组
    args_input = [pd_to_np(v) for v in args if is_np_pd(v)]  # 位置非数字参数
    args_param = [pd_to_np(v) for v in args if not is_np_pd(v)]  # 位置数字参数
    kwargs_input = {k: pd_to_np(v) for k, v in kwargs.items() if is_np_pd(v)}  # 命名非数字参数
    kwargs_param = {k: pd_to_np(v) for k, v in kwargs.items() if not is_np_pd(v)}  # 命名数字参数

    # 取一个非数字，得用于得到形状
    if len(args_input) > 0:
        real = args_input[0]
    else:
        real = list(kwargs_input.values())[0]

    if real.ndim == 1:
        # 一维，直接计算并返回
        return tafunc(*args_input, *args_param, **kwargs_input, **kwargs_param)

    # =====以下是二维======

    # 输出缓存
    outputs = [np.empty_like(real) for _ in output_names]
    # 输入源
    inputs = [*args_input, *kwargs_input.values()]

    with np.nditer(inputs + outputs,
                   flags=['external_loop'], order='F',
                   op_flags=[['readonly']] * len(inputs) + [['writeonly']] * len(outputs)) as it:
        for in_out in it:
            _in = in_out[:len(inputs)]  # 分离输入
            args_in = in_out[:len(args_input)]  # 分离位置输入
            kwargs_in = {k: v for k, v in zip(kwargs_input.keys(), _in[len(args_input):])}  # 分离命名输入

            # 计算并封装
            ta_out = tafunc(*args_in, *args_param, **kwargs_in, **kwargs_param)
            if not isinstance(ta_out, tuple):
                ta_out = tuple([ta_out])

            # 转存数据
            _out = in_out[len(inputs):]  # 分离输出
            for _i, _o in zip(_out, ta_out):
                _i[...] = _o

    # 输出
    if len(outputs) == 1:
        return outputs[0]
    return outputs


def _make_func(func, func_name, globals_):
    """复制一个函数，并命名"""
    import types as _types

    code_obj = func.__code__

    attr_list = ['co_argcount',
                 'co_kwonlyargcount',
                 'co_nlocals',
                 'co_stacksize',
                 'co_flags',
                 'co_code',
                 'co_consts',
                 'co_names',
                 'co_varnames',
                 'co_filename',
                 'co_name',
                 'co_firstlineno',
                 'co_lnotab']
    if _sys.version_info >= (3, 8):
        attr_list = ['co_argcount',
                     'co_posonlyargcount',
                     'co_kwonlyargcount',
                     'co_nlocals',
                     'co_stacksize',
                     'co_flags',
                     'co_code',
                     'co_consts',
                     'co_names',
                     'co_varnames',
                     'co_filename',
                     'co_name',
                     'co_firstlineno',
                     'co_lnotab']
    co_kwargs = {k[3:]: getattr(code_obj, k) for k in attr_list}
    co_kwargs['name'] = func_name  # 函数名在这里设置

    mycode_obj = _types.CodeType(*co_kwargs.values())
    return _types.FunctionType(mycode_obj, globals_, func_name, ())


def _ta_func(*args, **kwargs):
    """TA函数转发"""
    # 获取新函数名
    func_name = _sys._getframe().f_code.co_name

    # 获取指标信息
    info = _abstract.Function(func_name)._Function__info

    return tafunc_nditer(getattr(_talib, func_name),
                         args, kwargs,
                         info['output_names'])


for func_name in _talib.get_functions():
    """talib遍历"""
    _func = getattr(_talib, func_name)
    _f = _make_func(_ta_func, func_name, globals())

    _f.__doc__ = _func.__doc__
    globals()[func_name] = _f



import numpy as np

from .nb import _FILTER_1d_nb, _FILTER_2d_nb
from .reference import SUM
from .utils import pd_to_np, np_to_pd


def CROSS(S1, S2):
    """向上金叉"""
    arr = np.full_like(S1, fill_value=False, dtype=bool)
    arr[1:] = (S1 <= S2)[:-1] & (S1 > S2)[1:]
    return arr


def IF(condition, T, F):
    """序列布尔判断"""
    return np.where(condition, T, F)


def EVERY(real, timeperiod=5):
    """最近timeperiod天是否全为True"""
    return SUM(real, timeperiod) == timeperiod


def EXIST(real, timeperiod=5):
    """最近timeperiod天是否存在一天为True"""
    return SUM(real, timeperiod) > 0


def FILTER(S, N):
    """FILTER函数，S满足条件后，将其后N周期内的数据置为0"""
    S = pd_to_np(S, copy=True)
    if S.ndim == 2:
        _FILTER_2d_nb(S, N)
    else:
        _FILTER_1d_nb(S, N)
    return S


def BETWEEN(S, A, B):
    """BETWEEN(A,B,C)表示A处于B和C之间时返回1,否则返回0"""
    return ((A < S) & (S < B)) | ((A > S) & (S > B))


def VALUEWHEN(S, X):
    """"""
    return np_to_pd(np.where(S, X, np.nan)).ffill()



import bottleneck as bn
import numpy as np

from .utils import np_to_pd


def AVEDEV(real, timeperiod: int = 5):
    """平均绝对偏差"""

    def mad(x):
        return np.abs(x - x.mean()).mean()

    return np_to_pd(real).rolling(window=timeperiod).apply(mad, raw=True)


def STD(real, timeperiod: int = 5):
    """样本标准差"""
    return bn.move_std(real, window=timeperiod, axis=0, ddof=1)


def STDP(real, timeperiod: int = 5):
    """总体标准差"""
    return bn.move_std(real, window=timeperiod, axis=0, ddof=0)


def VAR(real, timeperiod: int = 5):
    """样本方差"""
    return bn.move_var(real, window=timeperiod, axis=0, ddof=1)


def VARP(real, timeperiod: int = 5):
    """总体方差"""
    return bn.move_std(real, window=timeperiod, axis=0, ddof=0)



import pandas as pd
import talib as ta
from talib import MA_Type

import ta_cn.talib as ta2d
from .ta import TA_SET_COMPATIBILITY, TA_COMPATIBILITY_METASTOCK
from .maths import MEAN
from .reference import COUNT, REF, DIFF, MA


def BBI(real, timeperiod1=3, timeperiod2=6, timeperiod3=12, timeperiod4=24):
    """BBI多空指标"""
    return MEAN(MA(real, timeperiod1),
                MA(real, timeperiod2),
                MA(real, timeperiod3),
                MA(real, timeperiod4))


def DPO(real, timeperiod=20):
    """DPO区间震荡线

    Parameters
    ----------
    real
    timeperiod:int
        常用参数：12

    Examples
    --------
    股票软件上常再求一次MA
    >>> MA(DPO(CLOSE, 20), 6)

    """
    return real - REF(MA(real, timeperiod), timeperiod // 2)


def MACD(real: pd.DataFrame, fastperiod=12, slowperiod=26, signalperiod=9):
    """MACD指标"""
    TA_SET_COMPATIBILITY(TA_COMPATIBILITY_METASTOCK)

    macd, macdsignal, macdhist = ta2d.MACDEXT(real,
                                              fastperiod=fastperiod, fastmatype=MA_Type.EMA,
                                              slowperiod=slowperiod, slowmatype=MA_Type.EMA,
                                              signalperiod=signalperiod, signalmatype=MA_Type.EMA)

    # macd起始位不是按slowperiod-1，而是按slowperiod+signalperiod-2，可能是为了三个输出的起始位相同
    # talib中的MACD没有*2
    return macd, macdsignal, macdhist * 2


def MTM(real, timeperiod=12):  # 动量指标
    """MTM动量指标

    Parameters
    ----------
    real
    timeperiod:int
        常用参数：12

    Examples
    --------
    股票软件上常再求一次MA
    >>> MA(MTM(CLOSE, 12), 6)

    """
    if real.ndim == 2:
        return DIFF(real, timeperiod)
    else:
        return ta.MOM(real, timeperiod=timeperiod)


def PSY(real, timeperiod=12):
    """PSY心理线

    Parameters
    ----------
    real
    timeperiod:int
        常用参数：12

    Examples
    --------
    股票软件上常再求一次MA
    >>> MA(PSY(CLOSE, 12), 6)

    """
    return COUNT(real > REF(real, 1), timeperiod) / timeperiod * 100


def DM(high, low, timeperiod=14):
    """Directional Movement方向动量

    WS_SUM威尔德平滑求和
    """
    return ta2d.PLUS_DM(high, low, timeperiod=timeperiod), ta2d.MINUS_DM(high, low, timeperiod=timeperiod)


def DI(high, low, close, timeperiod=14):
    """Directional Indicator方向指标"""
    return ta2d.PLUS_DI(high, low, close, timeperiod=timeperiod), ta2d.MINUS_DI(high, low, close, timeperiod=timeperiod)


def DMI(high, low, close, timeperiod=14):
    """趋向指标"""
    return (ta2d.PLUS_DI(high, low, close, timeperiod=timeperiod),
            ta2d.MINUS_DI(high, low, close, timeperiod=timeperiod),
            ta2d.ADX(high, low, close, timeperiod=timeperiod),
            ta2d.ADXR(high, low, close, timeperiod=timeperiod),
            )


def TRIX(real, timeperiod=12):
    """三重指数平滑均线

    由EMA算法差异导致的不同
    """
    TA_SET_COMPATIBILITY(TA_COMPATIBILITY_METASTOCK)

    return ta2d.TRIX(real, timeperiod=timeperiod)



import talib as ta

from .logical import IF
from .maths import SGN
from .reference import SUM, DIFF, REF


def OBV(real, volume, scale=1):  # 能量潮指标
    """能量潮指标"""
    S = SGN(DIFF(real))
    # 第一天由0改成1就与talib一样了
    S[0] = 1
    return SUM(S * volume, 0)


def OBV_CN(real, volume, scale=1 / 10000):
    """能量潮指标中国版"""
    S = SGN(DIFF(real))
    # 同花顺最后会除10000，但东方财富没有除
    return SUM(S * volume, 0) * scale


def VR(close, volume, timeperiod=26):
    """VR容量比率"""
    LC = REF(close, timeperiod=1)
    return SUM(IF(close > LC, volume, 0), timeperiod) / SUM(IF(close <= LC, volume, 0), timeperiod) * 100



from typing import Union

import numpy as np
import pandas as pd


def np_to_pd(x: np.ndarray,
             copy: bool = False,
             index=None,
             columns=None) -> Union[pd.DataFrame, pd.Series]:
    """将ndarray类型的数据转化为Pandas中的Series或DataFrame

    Parameters
    ----------
    x: ndarray
        源数据
    copy: bool
        是否需要先建立副本
    index: pd.Index
        创建pd.DataFrame/pd.Series的index参数
    columns: pd.Index
        创建pd.DataFrame的columns参数

    Returns
    -------
    pd.DataFrame or pd.Series
        根据维数转换后DataFrame或Series类型的数据

    Examples
    --------
    >>> df = np_to_pd(arr, copy=False, index=close.index, columns=close.columns)

    """
    if copy:
        x = x.copy()
    if isinstance(x, np.ndarray):
        if x.ndim == 2:
            return pd.DataFrame(x, index=index, columns=columns)
        elif x.ndim == 1:
            return pd.Series(x, index=index)
        assert False
    return x


def pd_to_np(x: Union[pd.DataFrame, pd.Series], copy: bool = False) -> np.ndarray:
    """将Pandas中的Series、DataFrame类型数据转化为ndarray

    Parameters
    ----------
    x: pd.DataFrame or pd.Series
        需要进行转换的数据
    copy: bool
        是否需要先建立副本

    Returns
    -------
    ndarray
        转换后ndarray类型的数据

    Examples
    --------
    >>> arr = pd_to_np(df, copy=False)

    """
    if copy:
        x = x.copy()

    if isinstance(x, (pd.Series, pd.DataFrame)):
        return x.values

    return x


def fillna(arr, direction='down'):
    """numpy版fillna

    Parameters
    ----------
    arr: np.darray
    direction: str
        up, down, left, right

    Examples
    --------
    >>> arr = np.array([[5, np.nan, np.nan, 7, 2],
                [3, np.nan, 1, 8, np.nan],
                [np.nan, 9, 6, np.nan, np.nan],
                [2, np.nan, 6, 2, np.nan],])
    >>> fillna(arr, direction='down')

    References
    ----------
    https://stackoverflow.com/questions/41190852/most-efficient-way-to-forward-fill-nan-values-in-numpy-array

    """
    mask = np.isnan(arr)
    if direction == 'down':
        row = np.where(~mask, np.arange(mask.shape[0])[:, None], 0)
        np.maximum.accumulate(row, axis=0, out=row)
        out = arr[row, np.arange(row.shape[1])]
    if direction == 'up':
        row = np.where(~mask, np.arange(mask.shape[0])[:, None], mask.shape[0] - 1)
        row = np.minimum.accumulate(row[::-1, :], axis=0)[::-1, :]
        out = arr[row, np.arange(row.shape[1])]
    if direction == 'left':
        col = np.where(~mask, np.arange(mask.shape[1]), mask.shape[1] - 1)
        col = np.minimum.accumulate(col[:, ::-1], axis=1)[:, ::-1]
        out = arr[np.arange(col.shape[0])[:, None], col]
    if direction == 'right':
        col = np.where(~mask, np.arange(mask.shape[1]), 0)
        np.maximum.accumulate(col, axis=1, out=col)
        out = arr[np.arange(col.shape[0])[:, None], col]

    return out


def zero_runs(a):
    """查找0的边界

    Parameters
    ----------
    a

    Examples
    --------
    >>> a = [1, 2, 3, 0, 0, 0, 0, 0, 0, 4, 5, 6, 0, 0, 0, 0, 9, 8, 7, 0, 10, 11, 0]
    >>> zero_runs(a)

    References
    ----------
    https://stackoverflow.com/questions/24885092/finding-the-consecutive-zeros-in-a-numpy-array/

    """
    # Create an array that is 1 where a is 0, and pad each end with an extra 0.
    iszero = np.concatenate(([0], np.equal(a, 0).view(np.int8), [0]))
    absdiff = np.abs(np.diff(iszero))
    # Runs start and end where absdiff is 1.
    ranges = np.where(absdiff == 1)[0].reshape(-1, 2)
    return ranges


def fill_zeros_with_last(arr):
    """0用上一个值填充，类似于ffill

    Parameters
    ----------
    arr

    Examples
    --------
    >>> arr = np.array([1, 0, 0, 8, 0, 4, 6, 2, 0, 0, 0, 0, 2])
    >>> fill_zeros_with_last(arr)

    References
    ----------
    https://stackoverflow.com/questions/30488961/fill-zero-values-of-1d-numpy-array-with-last-non-zero-values

    """
    prev = np.arange(len(arr))
    prev[arr == 0] = 0
    prev = np.maximum.accumulate(prev)
    return arr[prev]


def pushna(arr, direction='down'):
    """将非空数据按方向移动

    Parameters
    ----------
    arr: np.darray
    direction: str
        up, down, left, right

    Returns
    -------
    arr
        重排后的数据
    row
        行索引。将用于还原
    col
        列索引。将用于还原

    References
    ----------
    https://stackoverflow.com/questions/32062157/move-non-empty-cells-to-the-left-in-pandas-dataframe
    https://stackoverflow.com/questions/39361839/move-non-empty-cells-to-left-in-grouped-columns-pandas

    """
    if direction == 'down':
        row = (~np.isnan(arr)).argsort(axis=0, kind='stable')
        col = np.arange(arr.shape[1])[None]
    if direction == 'up':
        row = np.isnan(arr).argsort(axis=0, kind='stable')
        col = np.arange(arr.shape[1])[None]
    if direction == 'left':
        col = np.isnan(arr).argsort(axis=1, kind='stable')
        row = np.arange(arr.shape[0])[:, None]
    if direction == 'right':
        col = (~np.isnan(arr)).argsort(axis=1, kind='stable')
        row = np.arange(arr.shape[0])[:, None]

    return arr[row, col], row, col


def pullna(arr, row, col):
    """pushna的还原

    Parameters
    ----------
    arr: np.array
    row
        行索引
    col
        列索引

    Examples
    --------
    >>> a, row, col = pushna(df)
    >>> print(pullna(a, row, col))

    """
    tmp = np.empty_like(arr)
    tmp[row, col] = arr
    return tmp


if __name__ == '__main__':
    arr = np.array([[5, np.nan, np.nan, 7, 2],
                    [3, np.nan, 1, 8, np.nan],
                    [np.nan, 9, 6, np.nan, np.nan],
                    [2, np.nan, 6, 2, np.nan], ])
    print(fillna(arr, direction='down'))
    print(fillna(arr, direction='up'))
    print(fillna(arr, direction='left'))
    print(fillna(arr, direction='right'))

    print(pushna(arr, direction='down'))
    print(pushna(arr, direction='up'))
    print(pushna(arr, direction='left'))
    print(pushna(arr, direction='right'))



import pandas as pd

from .reference import MA
from .statistics import STDP


def BOLL(real: pd.DataFrame, timeperiod=20, nbdevup=2, nbdevdn=2):
    """BOLL指标，布林带

    使用总体标准差

    References
    ----------
    https://en.wikipedia.org/wiki/Bollinger_Bands

    """
    MID = MA(real, timeperiod)
    # 这里是总体标准差，值比样本标准差小。部分软件使用样本标准差是错误的，
    _std = STDP(real, timeperiod)
    UPPER = MID + _std * nbdevup
    LOWER = MID - _std * nbdevdn
    return UPPER, MID, LOWER



import talib as ta
from talib import MA_Type

import ta_cn.talib as ta2d
from .maths import MEAN
from .reference import LLV, HHV, REF, MA, TR
from .ta import TA_SET_COMPATIBILITY, TA_COMPATIBILITY_DEFAULT, TA_COMPATIBILITY_METASTOCK


def ATR_CN(high, low, close, timeperiod=14):
    """ATR真实波幅N日平均

    talib的ATR算法类似于EMA，所以要重写此处才与中国ATR相同
    """
    return MA(TR(high, low, close), timeperiod)


def BIAS(real, timeperiod=6):
    """BIAS乖离率

    Parameters
    ----------
    real
    timeperiod:int
        常用参数：6,12,24

    """
    return (real / MA(real, timeperiod) - 1) * 100


def KDJ(high, low, close, fastk_period=9, M1=3, M2=3):
    """KDJ指标

    talib中EMA的参数用法不同
    """
    TA_SET_COMPATIBILITY(TA_COMPATIBILITY_METASTOCK)

    K, D = ta2d.STOCH(high, low, close,
                      fastk_period=fastk_period,
                      slowk_period=(M1 * 2 - 1), slowk_matype=MA_Type.EMA,
                      slowd_period=(M2 * 2 - 1), slowd_matype=MA_Type.EMA)

    J = K * 3 - D * 2
    return K, D, J


def ROC(real, timeperiod=12):  # 变动率指标
    """ROC变动率指标

    Parameters
    ----------
    real
    timeperiod:int
        常用参数：12

    Examples
    --------
    股票软件上常再求一次MA
    >>> MA(ROC(CLOSE, 12), 6)

    """
    if real.ndim == 2:
        return (real / REF(real, timeperiod) - 1) * 100
    else:
        return ta.ROC(real, timeperiod=timeperiod)


def TYP(high, low, close):
    """典型价格。高低收的平均值"""
    if close.ndim == 2:
        return MEAN(high, low, close)
    else:
        return ta.TYPPRICE(high, low, close)


def WR(high, low, close, timeperiod=10):
    """W&R威廉指标

    Parameters
    ----------
    high
    low
    close
    timeperiod: int
        常用参数：10, 6

    Returns
    -------

    """

    if close.ndim == 2:
        hh = HHV(high, timeperiod)
        ll = LLV(low, timeperiod)
        return (hh - close) / (hh - ll) * 100
    else:
        return ta.WILLR(high, low, close, timeperiod=timeperiod)


def RSI(real, timeperiod=24):
    """RSI指标"""
    # 如果设置成1，将会多一个数字
    TA_SET_COMPATIBILITY(TA_COMPATIBILITY_DEFAULT)

    return ta2d.RSI(real, timeperiod=timeperiod)



import numpy as np
import pandas as pd

import ta_cn.talib as ta2d
from .ta import TA_SET_COMPATIBILITY, TA_COMPATIBILITY_DEFAULT, TA_COMPATIBILITY_METASTOCK
from .utils import np_to_pd

"""
由于MA有太多种了，单独提到此处，方便对比

默认EMA算法中，上一期值权重(timeperiod-1)/(timeperiod+1)，当前值权重2/(timeperiod+1)
ta.set_compatibility建议放在循环前执行

References
----------
https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.ewm.html?highlight=ewm#pandas.DataFrame.ewm
https://stackoverflow.com/questions/42869495/numpy-version-of-exponential-weighted-moving-average-equivalent-to-pandas-ewm

Warnings
--------
由于EMA的计算特点，只要调用了此文件中的公式，都需要预留一些数据，数据太短可能导致起点不同值不同

"""


def _MA_EMA(real, timeperiod, com=None, span=None, alpha=None):
    """内部函数。先计算MA做第一个值，然后再算EMA

    好几处开头的MA算法是一样的，但EMA的参数不同

    Notes
    -----
    二维矩阵如果开头有NaN，那么求均值的位置就不统一，得做特别处理

    """
    real = np_to_pd(real, copy=True)  # 开头部分将写入SMA
    ma = np_to_pd(np.zeros_like(real), copy=False)  # 来计算sma

    # 取最长位置, 用于计算SMA，没有必要全算一次MA
    max_end = real.notna().idxmax() + timeperiod
    if hasattr(max_end, 'max'):
        max_end = max_end.max()

    # 计算ma
    ma[:max_end] = real[:max_end].rolling(window=timeperiod, min_periods=timeperiod).mean()
    # 计算需要复制的区域,此区前部分为NaN,最后为mean
    mask = ma.isna().shift(fill_value=True)
    real[mask] = ma
    return real.ewm(com=com, span=span, alpha=alpha, min_periods=0, adjust=False).mean()


def EMA_0_TA(real, timeperiod=24):
    """EMA第一个值用MA"""
    TA_SET_COMPATIBILITY(TA_COMPATIBILITY_DEFAULT)
    return ta2d.EMA(real, timeperiod=timeperiod)


def EMA_0_PD(real, timeperiod=24):
    """EMA第一个值用MA"""
    return _MA_EMA(real, timeperiod=timeperiod, span=timeperiod)


def EMA_1_TA(real, timeperiod=24):
    """EMA第一个值用Price"""
    TA_SET_COMPATIBILITY(TA_COMPATIBILITY_METASTOCK)
    return ta2d.EMA(real, timeperiod=timeperiod)


def EMA_1_PD(real, timeperiod=24):
    """EMA第一个值用Price

    不少用到此指标，导致的计算过慢，如果原talib中有对应指标，并且与国内差别不大的情况下，可以迭代talib使用
    """
    return np_to_pd(real).ewm(span=timeperiod, min_periods=timeperiod, adjust=False).mean()


def SMA(real, timeperiod=24, M=1):
    """EMA第一个值用MA"""
    return _MA_EMA(real, timeperiod=timeperiod, alpha=M / timeperiod)


def DMA(real, alpha):
    """求X的动态移动平均。 0<alpha<1

    上一期值权重(1-alpha)，当前值权重alpha
    """
    return np_to_pd(real).ewm(alpha=alpha, adjust=False).mean()


def WS_SUM(real: pd.DataFrame, timeperiod: int = 5):
    """Wilder Smooth 威尔德平滑求和"""
    # 二维数据起始位置不同时，这里的算法会有问题
    # TODO: 还要再检查
    real = np_to_pd(real, copy=True)
    real[timeperiod - 1:timeperiod] = np.nansum(real[:timeperiod], axis=0)
    real[:timeperiod - 1] = 0

    return real.ewm(alpha=1 / timeperiod, min_periods=timeperiod).sum()

# if __name__ == '__main__':
#     import matplotlib.pyplot as plt
#
#     b = np.random.rand(10000).reshape(-1, 2)
#     b[:30, 0] = np.nan
#     b[:35, 1] = np.nan
#     a = b[:, 0]
#
#     r3 = WS_SUM(b).loc[:, 0]
#
#     r2 = EMA_0_PD(b).loc[:, 0]
#     r1 = EMA_0_TA(a)
#     pd.DataFrame({'DEFAULT_TALIB': r1, 'DEFAULT_PANDAS': r2}).plot()
#     pd.DataFrame({'DEFAULT_PANDAS': r2, 'DEFAULT_TALIB': r1}).plot()
#     plt.show()
#
#     r1 = EMA_1_TA(a)
#     r2 = EMA_1_PD(a)
#     pd.DataFrame({'METASTOCK_TALIB': r1, 'METASTOCK_PANDAS': r2}).plot()
#     pd.DataFrame({'METASTOCK_PANDAS': r2, 'METASTOCK_TALIB': r1}).plot()
#     plt.show()



import bottleneck as bn
import numpy as np

from .maths import MAX, ABS


def CONST(real):
    """取A最后的值为常量"""
    return np.full_like(real, real[-1])


def COUNT(real, timeperiod):
    """统计N周期中满足X条件的周期数"""
    return SUM(real, timeperiod)


def DIFF(real, timeperiod: int = 1):
    """差分"""
    return real - REF(real, timeperiod)


def HHV(real, timeperiod: int = 5):
    """求timeperiod周期内real最高值"""
    return bn.move_max(real, window=timeperiod, axis=0)


def HHVBARS(real, timeperiod: int = 5):
    """求timeperiod周期内real最高值到当前周期数"""
    return bn.move_argmax(real, window=timeperiod, axis=0)


def LLV(real, timeperiod: int = 5):
    """求timeperiod周期内real最低值"""
    return bn.move_min(real, window=timeperiod, axis=0)


def LLVBARS(real, timeperiod: int = 5):
    """求timeperiod周期内real最低值到当前周期数"""
    return bn.move_argmin(real, window=timeperiod, axis=0)


def REF(real, timeperiod: int = 1):
    """向前引用"""
    arr = np.full(real.shape, np.nan)
    arr[timeperiod:] = real[:-timeperiod]
    return arr


def SUM(real, timeperiod: int = 5):
    """时序滚动求和

    Parameters
    ----------
    real
    timeperiod: int
        滚动周期
        0表示全量即累计求和

    """
    if timeperiod > 0:
        return bn.move_sum(real, window=timeperiod, axis=0)
    else:
        # 0表示全量累计求和
        return np.nancumsum(real, axis=0)


def TR(high, low, close):
    """TR真实波幅"""
    lc = REF(close, 1)
    return MAX(high - low, ABS(high - lc), ABS(lc - low))


def MA(real, timeperiod: int = 5):
    """简单移动平均

    等价于talib中的SMA"""
    return bn.move_mean(real, window=timeperiod, axis=0)



from functools import reduce

import numpy as np


def ABS(x):
    """绝对值"""
    return np.abs(x)


def LN(x):
    """底是e的对数

    0返回-inf。不好看，还是先用where过滤
    """
    return np.log(np.where(x > 0, x, np.nan))


def LOG(x):
    """底是10的对数"""
    return np.log10(np.where(x > 0, x, np.nan))


def POW(x1, x2):
    """次方"""
    return np.power(x1, x2)


def REVERSE(x):
    """相反数"""
    return -x


def SQRT(x):
    """平方根"""
    with np.errstate(invalid='ignore'):
        return np.sqrt(x)


def MAX(*args):
    """求max

    Examples
    --------
    >>> MAX(1, 2)
    >>> MAX(1, 2, 3)

    """
    return reduce(np.maximum, args)


def MIN(*args):
    """求min"""
    return reduce(np.minimum, args)


def ADD(*args):
    """求和

    Examples
    --------
    >>> ADD(1, 2)
    >>> ADD(1, 2, 3)

    """
    return reduce(np.add, args)


def SUB(*args):
    """相减"""
    return reduce(np.subtract, args)


def MUL(*args):
    """连乘"""
    return reduce(np.multiply, args)


def DIV(*args):
    """连除

    RuntimeWarning: invalid value encountered in true_divide
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        return reduce(np.true_divide, args)


def MEAN(*args):
    """均值"""
    return ADD(*args) / len(args)


def ROUND(a, decimals=3):
    """四舍五入取3位小数"""
    return np.round(a, decimals)


def SGN(x):
    """求符号方向

    return (0 < x) * 1 - (x < 0)
    """
    return np.sign(x)


if __name__ == '__main__':
    a = np.random.rand(10000)
    a[:20] = np.nan
    a[:10] = -10
    a[:5] = 0
    b = ABS(a)
    print(b)
    b = LN(a)
    print(b)
    b = LOG(a)
    print(b)
    b = SQRT(a)
    print(b)
    b = DIV(a, a)
    print(b)
    b = SGN(a)
    print(b)



import setuptools

with open("README.md", "r", encoding='utf-8') as fp:
    long_description = fp.read()

version = {}
with open("ta_cn/_version.py", encoding="utf-8") as fp:
    exec(fp.read(), version)

setuptools.setup(
    name="ta_cn",
    version=version['__version__'],
    author="wukan",
    author_email="wu-kan@163.com",
    description="Technical Analysis Indicators",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wukan1986/ta_cn",
    packages=setuptools.find_packages(),
    install_requires=[
        'TA-Lib',
        'numpy',
    ],
    extras_require={
        'all': [
            # 'numpy>=1.20.0',  # 主要是为了sliding_window_view
            'pandas',
            'bottleneck',
            'TA-Lib>=0.4.19',
            'numba',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
    ],
    python_requires=">=3.7",
)
